<?php

namespace Drupal\monarch_ql\Controller;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\Sql\TableMappingInterface;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Session\AccountInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;

/**
 * Returns responses for QL routes.
 */
final class QlController extends ControllerBase {

  /**
   * The field default filter list.
   */
  private array $queryFieldFilters = [
    'file.status !=' => '0',
    'file.uri NOT LIKE' => 'private://%',
    'media.status !=' => '0',
    'node.status !=' => '0',
    'taxonomy_term.status !=' => '0',
  ];

  /**
   * Local cache for repeated actions.
   */
  private array $cache = [];

  /**
   * The entity_field.manager service.
   */
  private readonly EntityFieldManagerInterface $entityFieldManager;

  /**
   * The database service.
   */
  private readonly Connection $database;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    $instance = new static();

    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->entityFieldManager = $container->get('entity_field.manager');
    $instance->database = $container->get('database');
    $instance->moduleHandler = $container->get('module_handler');

    return $instance;
  }

  /**
   * Check access.
   *
   * @todo Implement hooks.
   */
  public function access(AccountInterface $account, $query = NULL, $page = 0, $count = 100, Request $request = NULL) : AccessResultInterface {
    $access = NULL;

    foreach (($this->moduleHandler->invokeAll('monarch_ql_access', [$account, $query, $page, $count, $request]) ?: []) as $v) {
      $access = ($access ?? $v) && $v;
    }

    return AccessResult::allowedIf($access ?? FALSE);
  }

  /**
   * Check entity type and field availability.
   */
  private function checkAvailable(array $field_access_list, string $entity_type, string $field_name = NULL) : bool {
    if ($field_name) {
      return (bool) ($field_access_list["$entity_type.$field_name"] ?? $field_access_list["$entity_type"] ?? NULL);
    }

    return (bool) ($field_access_list["$entity_type"] ?? NULL);
  }

  /**
   * Checks whether the given field can be stored in a shared table.
   *
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $storage_definition
   *   The field storage definition.
   *
   * @return bool
   *   TRUE if the field can be stored in a shared table, FALSE otherwise.
   */
  public function allowsSharedTableStorage(FieldStorageDefinitionInterface $storage_definition) {
    return !$storage_definition->hasCustomStorage() && $storage_definition->isBaseField() && !$storage_definition->isMultiple() && !$storage_definition->isDeleted();
  }

  /**
   * Checks whether the given field has to be stored in a dedicated table.
   *
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $storage_definition
   *   The field storage definition.
   *
   * @return bool
   *   TRUE if the field has to be stored in a dedicated table, FALSE otherwise.
   */
  public function requiresDedicatedTableStorage(FieldStorageDefinitionInterface $storage_definition) {
    return !$storage_definition->hasCustomStorage() && !$this->allowsSharedTableStorage($storage_definition);
  }

  /**
   * Get columns from a field storage definition.
   */
  public function getColumnNames(FieldStorageDefinitionInterface $storage_definition) : array {
    $_entity_type_id = $storage_definition->getTargetEntityTypeId();
    $_field_name = $storage_definition->getName();

    if (isset($this->cache[__FUNCTION__][$_entity_type_id][$_field_name])) {
      return $this->cache[__FUNCTION__][$_entity_type_id][$_field_name];
    }

    $ret = [];

    $field_name = $storage_definition->getName();

    if ($this->allowsSharedTableStorage($storage_definition)) {
      foreach (array_keys($storage_definition->getColumns()) as $property_name) {
        $ret[$property_name] = count($storage_definition->getColumns()) == 1 ? $field_name : $field_name . '__' . $property_name;
      }
    }
    elseif ($this->requiresDedicatedTableStorage($storage_definition)) {
      foreach (array_keys($storage_definition->getColumns()) as $property_name) {
        if ($property_name === TableMappingInterface::DELTA) {
          $ret[$property_name] = 'delta';
        }
        elseif ($property_name === 'deleted') {
          $ret[$property_name] = $property_name;
        }
        else {
          $ret[$property_name] = $field_name . '_' . $property_name;
        }
      }
    }

    return $this->cache[__FUNCTION__][$_entity_type_id][$_field_name] = $ret;
  }

  /**
   * Generates a safe and unambiguous field table name.
   *
   * The method accounts for a maximum table name length of 64 characters, and
   * takes care of disambiguation.
   *
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $storage_definition
   *   The field storage definition.
   *
   * @return string
   *   The final table name.
   */
  private function getDedicatedDataTableName(FieldStorageDefinitionInterface $storage_definition) {
    $_entity_type_id = $storage_definition->getTargetEntityTypeId();
    $_field_name = $storage_definition->getName();

    if (isset($this->cache[__FUNCTION__][$_entity_type_id][$_field_name])) {
      return $this->cache[__FUNCTION__][$_entity_type_id][$_field_name];
    }

    $entity_type_id = substr($storage_definition->getTargetEntityTypeId(), 0, EntityTypeInterface::ID_MAX_LENGTH);
    $table_name = $entity_type_id . '__' . $storage_definition->getName();

    if (strlen($table_name) > 48) {
      $field_hash = substr(hash('sha256', $storage_definition->getUniqueStorageIdentifier()), 0, 10);
      $table_name = $entity_type_id . '__' . $field_hash;

      if (strlen($table_name) > 48) {
        $table_name = '__' . $field_hash;
      }
    }

    return $this->cache[__FUNCTION__][$_entity_type_id][$_field_name] = $table_name;
  }

  /**
   * Gets the base fields for an entity type.
   */
  private function getBaseFields(ContentEntityTypeInterface $entity_type) : array {
    $ret = array_flip(array_filter([
      'id' => $entity_type->getKey('id'),
      'revision' => $entity_type->getKey('revision'),
      'bundle' => $entity_type->getKey('bundle'),
      'uuid' => $entity_type->getKey('uuid'),
      'langcode' => $entity_type->getKey('langcode'),
    ]));

    $base_table = $entity_type->getBaseTable();

    return array_map(function () use ($base_table) {
      return $base_table;
    }, $ret);
  }

  /**
   * Gets the data fields for an entity type.
   */
  private function getDataFields(ContentEntityTypeInterface $entity_type) : array {
    $ret = $this->entityFieldManager->getFieldStorageDefinitions($entity_type->id());

    foreach ($this->getBaseFields($entity_type) as $column => $key) {
      unset($ret[$column]);
    }

    foreach ($entity_type->getRevisionMetadataKeys() as $key => $column) {
      unset($ret[$column]);
    }

    foreach ($ret as $column => $field_storage_definition) {
      if ($this->allowsSharedTableStorage($field_storage_definition)) {
        $ret[$column] = $entity_type->getDataTable();
      }
      else {
        unset($ret[$column]);
      }
    }

    return $ret;
  }

  /**
   * Gets the additional fields.
   */
  private function getAdditionalFields(ContentEntityTypeInterface $entity_type) : array {
    $ret = $this->entityFieldManager->getFieldStorageDefinitions($entity_type->id());

    foreach ($ret as $field_name => $field_storage_definition) {
      if ($this->requiresDedicatedTableStorage($field_storage_definition)) {
        $ret[$field_name] = $this->getDedicatedDataTableName($field_storage_definition);
      }
      else {
        unset($ret[$field_name]);
      }
    }

    return $ret;
  }

  /**
   * Get the field table map.
   */
  private function getFieldTableMap(ContentEntityTypeInterface $entity_type) : array {
    $_entity_type_id = $entity_type->id();

    if (isset($this->cache[__FUNCTION__][$_entity_type_id])) {
      return $this->cache[__FUNCTION__][$_entity_type_id];
    }

    return $this->cache[__FUNCTION__][$_entity_type_id] = ($this->getBaseFields($entity_type) + $this->getDataFields($entity_type) + $this->getAdditionalFields($entity_type));
  }

  /**
   * Convert hex keys back to the original keys.
   */
  private function convertKeysFromHex(array $q) : array {
    $ret = [];

    foreach ($q as $key => $value) {
      $ret[urldecode(hex2bin($key))] = is_array($value) ? $this->convertKeysFromHex($value) : $value;
    }

    return $ret;
  }

  /**
   * Parses a query string trying to preserve period and space in key names.
   */
  private function parseQueryString(string $str) : array {
    $parts = explode('&', $str);

    foreach ($parts as $index => $part) {
      $subparts = explode('=', $part);

      if (!is_null($subparts[0])) {
        $p = preg_split('/([\[\]])/', $subparts[0], -1, PREG_SPLIT_DELIM_CAPTURE);

        foreach ($p as $i => $s) {
          if ($s !== '[' && $s !== ']') {
            $p[$i] = bin2hex($s);
          }
        }

        $subparts[0] = implode('', $p);
      }

      $parts[$index] = implode('=', $subparts);
    }

    parse_str(implode('&', $parts), $q);

    return $this->convertKeysFromHex($q);
  }

  /**
   * Flatten array.
   */
  private function flattenWithoutKeys(array $arr, array &$ret = []) : array {
    foreach ($arr as $value) {
      if (is_array($value)) {
        $this->flattenWithoutKeys($value, $ret);
      }
      else {
        $ret[] = $value;
      }
    }

    return $ret;
  }

  /**
   * Builds the response.
   */
  public function build($query = NULL, $page = 0, $count = 100, Request $request = NULL) : Response {
    if (is_null($query)) {
      return new Response(NULL, 404);
    }

    if (is_numeric($page)) {
      $page = max(0, intval($page));
    }
    else {
      $page = 0;
    }

    if (is_numeric($count)) {
      $count = min(100, max(1, intval($count)));
    }
    else {
      $count = 100;
    }

    $start = hrtime(TRUE);

    $field_access_list = [];

    $filter_list = $this->queryFieldFilters;
    $user_filter_list = $this->parseQueryString(explode('?', $request->getRequestUri(), 2)[1] ?? '');

    foreach ($user_filter_list as $key => $value) {
      if (is_array($value)) {
        $user_filter_list[$key] = $this->flattenWithoutKeys($value);
      }
    }

    $this->moduleHandler->alter('monarch_ql_field_access', $field_access_list);
    $this->moduleHandler->alter('monarch_ql_field_filters', $filter_list);

    $definitions = array_filter($this->entityTypeManager->getDefinitions(), function ($type) {
      return $type instanceof ContentEntityTypeInterface;
    });

    $query_list = [];

    foreach (explode(' ', $query) as $query_part) {
      [$entity_type_id, $field_name, $column] = explode('.', $query_part, 3) + [NULL, NULL, NULL];

      $definition = $definitions[$entity_type_id] ?? NULL;

      if (!$definition || !$definition->getBaseTable() || !($definition instanceof ContentEntityTypeInterface)) {
        continue;
      }

      $field_table_map = $this->getFieldTableMap($definition);

      $field_storage_definitions = $this->entityFieldManager->getFieldStorageDefinitions($entity_type_id);
      $id_key = $definition->getKey('id');

      if ($field_name) {
        if ($this->checkAvailable($field_access_list, $entity_type_id, $field_name)) {
          $query_list[$entity_type_id][$definition->getBaseTable()] = $query_list[$entity_type_id][$definition->getBaseTable()] ?? [];
        }
        else {
          continue;
        }

        $query_list[$entity_type_id][$definition->getBaseTable()] = $query_list[$entity_type_id][$definition->getBaseTable()] ?? [];

        if ($field_storage_definition = $field_storage_definitions[$field_name] ?? NULL) {
          $columns = $this->getColumnNames($field_storage_definition);
          $table_name = $field_table_map[$field_name] ?? NULL;

          if (!$table_name) {
            continue;
          }

          if ($column) {
            if ($columns[$column] ?? NULL) {
              $query_list[$entity_type_id][$table_name][$field_name][$column] = $columns[$column];
            }
          }
          else {
            foreach (array_keys($columns) as $column) {
              $query_list[$entity_type_id][$table_name][$field_name][$column] = $columns[$column];
            }
          }
        }
      }
      else {
        foreach ($field_storage_definitions as $field_name => $field_storage_definition) {
          if ($this->checkAvailable($field_access_list, $entity_type_id, $field_name)) {
            $query_list[$entity_type_id][$definition->getBaseTable()] = $query_list[$entity_type_id][$definition->getBaseTable()] ?? [];
          }
          else {
            continue;
          }

          $table_name = $field_table_map[$field_name] ?? NULL;

          if (!$table_name) {
            continue;
          }

          $columns = $this->getColumnNames($field_storage_definition);

          foreach (array_keys($columns) as $column) {
            $query_list[$entity_type_id][$table_name][$field_name][$column] = $columns[$column];
          }
        }
      }
    }

    $combined_filter_list = [
      ...$filter_list,
      ...$user_filter_list,
    ];

    $filter_list = [];

    foreach ($combined_filter_list as $filter => $value) {
      [$filter, $operator] = explode(' ', $filter, 2) + [NULL, NULL];
      [$entity_type_id, $field_name, $column] = explode('.', $filter, 3) + [NULL, NULL, NULL];

      if (!isset($query_list[$entity_type_id]) || empty($field_name)) {
        continue;
      }

      if (!$this->checkAvailable($field_access_list, $entity_type_id, $field_name)) {
        continue;
      }

      $field = $this->entityFieldManager->getFieldStorageDefinitions($entity_type_id)[$field_name] ?? NULL;

      if (!$field) {
        continue;
      }

      if (empty($column)) {
        $column = $field->getMainPropertyName();
      }

      $columns = $this->getColumnNames($field);
      $column = $columns[$column] ?? NULL;

      if (!$column) {
        continue;
      }

      $table_name = $field_table_map[$field_name] ?? NULL;

      if (!$table_name) {
        continue;
      }

      if ($operator) {
        $filter_list[$entity_type_id][$table_name][$column][] = [$value, $operator];
      }
      else {
        if (is_array($value)) {
          $filter_list[$entity_type_id][$table_name][$column][] = [$value, 'IN'];
        }
        else {
          $filter_list[$entity_type_id][$table_name][$column][] = [$value, '='];
        }
      }
    }

    $json = [];

    foreach ($query_list as $entity_type_id => $query_tables) {
      $definition = $definitions[$entity_type_id];
      $id_key = $definition->getKey('id');

      $ids = NULL;

      foreach ($query_tables as $query_table => $query_fields) {
        $query = $this->database->select($query_table, 't');
        $real_id_key = $query_table === $definition->getBaseTable() || $query_table === $definition->getDataTable() ? $id_key : 'entity_id';
        $query->addField('t', $real_id_key);

        foreach ($query_fields as $query_field => $columns) {
          foreach ($columns as $column) {
            if ($column !== $real_id_key) {
              $query->addField('t', $column);
            }
          }
        }

        if ($query_table === $definition->getBaseTable()) {
          foreach ($filter_list[$entity_type_id] ?? [] as $filter_table => $filter_columns) {
            $filter_query = $this->database->select($filter_table, 'f');
            $filter_id_key = $filter_table === $definition->getBaseTable() || $filter_table === $definition->getDataTable() ? $id_key : 'entity_id';
            $filter_query->addField('f', $filter_id_key, 'id');

            foreach ($filter_columns as $filter_column => $filter_params_list) {
              foreach ($filter_params_list as $filter_params) {
                $filter_query->condition($filter_column, ...$filter_params);
              }
            }

            $query->condition($real_id_key, $filter_query, 'IN');
          }

          foreach ($query->orderBy($id_key)->range($page * $count, $count)->execute() as $row) {
            foreach ($row as $column => $value) {
              if (is_null($value)) {
                continue;
              }

              $json[$entity_type_id][$row->{$real_id_key}][$column] = $value;
            }

            $ids[] = $row->{$real_id_key};
          }
        }
        elseif ($ids) {
          if ($query_table === $definition->getDataTable()) {
            foreach ($query->condition($real_id_key, $ids, 'IN')->execute() as $row) {
              foreach ($row as $column => $value) {
                if (is_null($value)) {
                  continue;
                }

                $json[$entity_type_id][$row->{$real_id_key}][$column] = $value;
              }
            }
          }
          else {
            $query->addField('t', 'delta');

            $column_lookup = array_flip($columns);

            foreach ($query->condition($real_id_key, $ids, 'IN')->execute() as $row) {
              foreach ($row as $key => $value) {
                if (($column_lookup[$key] ?? NULL) && !is_null($value)) {
                  $json[$entity_type_id][$row->{$real_id_key}][$query_field][$row->delta][$column_lookup[$key]] = $value;
                }
              }
            }
          }
        }
      }

      if (is_array($json[$entity_type_id] ?? NULL)) {
        $json[$entity_type_id] = array_values($json[$entity_type_id]);
      }
    }

    return new JsonResponse((object) $json, 200, [
      'x-elapsed' => (hrtime(TRUE) - $start) / 1000000.0,
    ]);
  }

}
